from HABApp.core.events import ValueChangeEventFilter, ValueUpdateEventFilter
from . import MqttValueChangeEvent, MqttValueUpdateEvent


class MqttValueUpdateEventFilter(ValueUpdateEventFilter):
    _EVENT_TYPE = MqttValueUpdateEvent


class MqttValueChangeEventFilter(ValueChangeEventFilter):
    _EVENT_TYPE = MqttValueChangeEvent



from pathlib import Path

import HABApp
from HABApp.rule_manager import RuleFile
from .bench_habapp import HABAppBenchRule
from .bench_oh import OpenhabBenchRule
from .bench_mqtt import MqttBenchRule


class BenchFile(RuleFile):
    def __init__(self, rule_manager):
        super().__init__(rule_manager, path=Path('BenchmarkFile'))

    def create_rules(self, created_rules: list):
        glob = globals()
        glob['__HABAPP__RUNTIME__'] = self.rule_manager.runtime
        glob['__HABAPP__RULE_FILE__'] = self
        glob['__HABAPP__RULES'] = created_rules

        rule_ha = rule = HABAppBenchRule()
        if HABApp.CONFIG.mqtt.connection.host:
            rule = rule.link_rule(MqttBenchRule())
        if HABApp.CONFIG.openhab.connection.host:
            rule = rule.link_rule(OpenhabBenchRule())

        rule_ha.run_in(5, rule_ha.do_bench_start)

        glob.pop('__HABAPP__RUNTIME__')
        glob.pop('__HABAPP__RULE_FILE__')
        glob.pop('__HABAPP__RULES')



from typing import Any

import HABApp
from HABApp.core.const import MISSING
from . import ValueChangeEvent, ValueUpdateEvent


class EventFilter:
    def __init__(self, event_type, **kwargs):
        assert len(kwargs) < 3, 'EventFilter only allows up to two args that will be used to filter'

        for arg in kwargs:
            if arg not in event_type.__annotations__:
                raise AttributeError(f'Filter attribute "{arg}" does not exist for "{event_type.__name__}"')

        self.__cls = event_type
        self.__filter = kwargs

    def create_event_listener(self, name, cb) -> 'HABApp.core.EventBusListener':
        kwargs = {'event_type': self.__cls}
        ct = 1
        for k, v in self.__filter.items():
            kwargs[f'attr_name{ct}'] = k
            kwargs[f'attr_value{ct}'] = v
            ct += 1

        return HABApp.core.EventBusListener(name, cb, **kwargs)

    def __repr__(self):
        name = self.__class__.__name__
        vals = [f'{k}={v}' for k, v in self.__filter.items()]
        if name == EventFilter.__name__:
            vals.insert(0, f'event_type={self.__cls.__name__}')
        return f'{name}({", ".join(vals)})'


class ValueUpdateEventFilter(EventFilter):
    _EVENT_TYPE = ValueUpdateEvent

    def __init__(self, value):
        super().__init__(self._EVENT_TYPE, value=value)


class ValueChangeEventFilter(EventFilter):
    _EVENT_TYPE = ValueChangeEvent

    def __init__(self, value: Any = MISSING, old_value: Any = MISSING):
        args = {}
        if value is not MISSING:
            args['value'] = value
        if old_value is not MISSING:
            args['old_value'] = old_value
        super().__init__(self._EVENT_TYPE, **args)



from .events import ComplexEventValue, ValueUpdateEvent, ValueChangeEvent, \
    ItemNoChangeEvent, ItemNoUpdateEvent, AllEvents
from . import habapp_events
from .event_filters import EventFilter, ValueChangeEventFilter, ValueUpdateEventFilter



import itertools
import logging
import logging.handlers
import traceback
import typing
from asyncio import iscoroutinefunction, run_coroutine_threadsafe, sleep
from dataclasses import dataclass
from types import FunctionType, MethodType
from typing import Callable, Coroutine, Union

from HABApp.core.const import loop


@dataclass(frozen=True)
class ShutdownInfo:
    func: Union[Callable[[], typing.Any], Coroutine]
    msg: str
    last: bool


_FUNCS: typing.List[ShutdownInfo] = []

requested: bool = False


def register_func(func, last=False, msg: str = ''):
    assert isinstance(func, (FunctionType, MethodType)) or iscoroutinefunction(func), print(type(func))
    assert last is True or last is False, last
    assert isinstance(msg, str)

    _FUNCS.append(ShutdownInfo(func, f'{func.__module__}.{func.__name__}' if not msg else msg, last))


def request_shutdown():
    run_coroutine_threadsafe(_shutdown(), loop)


async def _shutdown():
    global requested
    "Request execution of all functions"

    log = logging.getLogger('HABApp.Shutdown')
    log.debug('Requested shutdown')

    requested = True

    for obj in itertools.chain(filter(lambda x: not x.last, _FUNCS),
                               filter(lambda x: x.last, _FUNCS)):
        try:
            log.debug(f'{obj.msg}')
            if iscoroutinefunction(obj.func):
                await obj.func()
            else:
                obj.func()
            log.debug('-> done!')
            await sleep(0.02)
        except Exception as ex:
            log.error(ex)
            tb = traceback.format_exc().splitlines()
            for line in tb:
                log.error(line)

    log.debug('Shutdown complete')
    return None



import logging
import typing

import paho.mqtt.client as mqtt

import HABApp
from HABApp.core import Items
from HABApp.core.wrapper import log_exception
from HABApp.runtime import shutdown
from .events import MqttValueChangeEvent, MqttValueUpdateEvent
from ..core.const.json import load_json

log = logging.getLogger('HABApp.mqtt.connection')
log_msg = logging.getLogger('HABApp.EventBus.mqtt')


class MqttStatus:
    def __init__(self):
        self.loop_started = False
        self.connected = False
        self.client: typing.Optional[mqtt.Client] = None
        self.subscriptions: typing.List[typing.Tuple[str, int]] = []


STATUS = MqttStatus()


def setup():
    config = HABApp.config.CONFIG.mqtt

    # config changes
    config.subscribe.subscribe_for_changes(subscription_changed)
    config.connection.subscribe_for_changes(connect)

    # shutdown
    shutdown.register_func(disconnect, msg='Disconnecting MQTT')


def connect():
    config = HABApp.config.CONFIG.mqtt

    if not config.connection.host:
        log.info('MQTT disabled')
        disconnect()
        return None

    if STATUS.connected:
        log.info('disconnecting')
        STATUS.client.disconnect()
        STATUS.connected = False

    STATUS.client = mqtt.Client(
        client_id=config.connection.client_id,
        clean_session=False
    )

    if config.connection.tls:
        STATUS.client.tls_set()

        # we can only set tls_insecure if we have a tls connection
        if config.connection.tls_insecure:
            log.warning('Verification of server hostname in server certificate disabled!')
            log.warning('Use this only for testing, not for a real system!')
            STATUS.client.tls_insecure_set(True)

    # set user/pw if required
    user = config.connection.user
    pw = config.connection.password
    if user:
        STATUS.client.username_pw_set(user, pw if pw else None)

    # setup callbacks
    STATUS.client.on_connect = on_connect
    STATUS.client.on_disconnect = on_disconnect
    STATUS.client.on_message = process_msg

    STATUS.client.connect_async(
        config.connection.host, port=config.connection.port, keepalive=60
    )

    log.info(f'Connecting to {config.connection.host}:{config.connection.port}')

    if not STATUS.loop_started:
        STATUS.client.loop_start()
    STATUS.loop_started = True


@log_exception
def disconnect():
    if STATUS.connected:
        STATUS.client.disconnect()
        STATUS.connected = False
    if STATUS.loop_started:
        STATUS.client.loop_stop()
        STATUS.loop_started = False

    STATUS.client = None


@log_exception
def on_connect(client, userdata, flags, rc):
    log.log(logging.INFO if not rc else logging.ERROR, mqtt.connack_string(rc))
    if rc:
        return None
    STATUS.connected = True

    STATUS.subscriptions.clear()
    subscription_changed()


@log_exception
def on_disconnect(client, userdata, rc):
    log.log(logging.INFO if not rc else logging.ERROR, f'Disconnect: {mqtt.error_string(rc)} ({rc})')
    STATUS.connected = False


@log_exception
def subscription_changed():
    if not STATUS.connected:
        return None

    if STATUS.subscriptions:
        unsubscribe = [k[0] for k in STATUS.subscriptions]
        log.debug('Unsubscribing from:')
        for t in unsubscribe:
            log.debug(f' - "{t}"')
        STATUS.client.unsubscribe(unsubscribe)

    topics = HABApp.config.CONFIG.mqtt.subscribe.topics
    default_qos = HABApp.config.CONFIG.mqtt.subscribe.qos
    STATUS.subscriptions = [(topic, qos if qos is not None else default_qos) for topic, qos in topics]
    log.debug('Subscribing to:')
    for topic, qos in STATUS.subscriptions:
        log.debug(f' - "{topic}" (QoS {qos:d})')
    STATUS.client.subscribe(STATUS.subscriptions)


@log_exception
def process_msg(client, userdata, message: mqtt.MQTTMessage):
    topic: str = message.topic

    try:
        payload = message.payload.decode("utf-8")

        if log_msg.isEnabledFor(logging.DEBUG):
            log_msg._log(logging.DEBUG, f'{topic} ({message.qos}): {payload}', [])

        # load json dict and list
        if payload.startswith('{') and payload.endswith('}') or payload.startswith('[') and payload.endswith(']'):
            try:
                payload = load_json(payload)
            except ValueError:
                pass
        else:
            # try to cast to int/float
            try:
                payload = int(payload)
            except ValueError:
                try:
                    payload = float(payload)
                except ValueError:
                    pass
    except UnicodeDecodeError:
        # Payload ist binary
        payload = message.payload
        if log_msg.isEnabledFor(logging.DEBUG):
            log_msg._log(logging.DEBUG, f'{topic} ({message.qos}): {payload[:20]}...', [])

    _item = None    # type: typing.Optional[HABApp.mqtt.items.MqttBaseItem]
    try:
        _item = Items.get_item(topic)   # type: HABApp.mqtt.items.MqttBaseItem
    except HABApp.core.Items.ItemNotFoundException:
        # only create items for if the message has the retain flag
        if message.retain:
            _item = Items.create_item(topic, HABApp.mqtt.items.MqttItem)  # type: HABApp.mqtt.items.MqttItem

    # we don't have an item -> we process only the event
    if _item is None:
        HABApp.core.EventBus.post_event(topic, MqttValueUpdateEvent(topic, payload))
        return None

    # Remember state and update item before doing callbacks
    _old_state = _item.value
    _item.set_value(payload)

    # Post events
    HABApp.core.EventBus.post_event(topic, MqttValueUpdateEvent(topic, payload))
    if _old_state != payload:
        HABApp.core.EventBus.post_event(topic, MqttValueChangeEvent(topic, payload, _old_state))



from HABApp.core.events import ValueChangeEventFilter, ValueUpdateEventFilter
from . import ItemStateChangedEvent, ItemStateEvent


class ItemStateEventFilter(ValueUpdateEventFilter):
    _EVENT_TYPE = ItemStateEvent


class ItemStateChangedEventFilter(ValueChangeEventFilter):
    _EVENT_TYPE = ItemStateChangedEvent
